@@ -355,6 +355,7 @@ gspell_checker_check_word (GspellChecker  *checker,
 	GspellCheckerPrivate *priv;
 	gint enchant_result;
 	gboolean correctly_spelled;
+	gchar *sanitized_word;
 
 	g_return_val_if_fail (GSPELL_IS_CHECKER (checker), FALSE);
 	g_return_val_if_fail (word != NULL, FALSE);
@@ -373,7 +374,15 @@ gspell_checker_check_word (GspellChecker  *checker,
 		return TRUE;
 	}
 
-	enchant_result = enchant_dict_check (priv->dict, word, word_length);
+	if (_gspell_utils_str_to_ascii_apostrophe (word, word_length, &sanitized_word))
+	{
+		enchant_result = enchant_dict_check (priv->dict, sanitized_word, -1);
+		g_free (sanitized_word);
+	}
+	else
+	{
+		enchant_result = enchant_dict_check (priv->dict, word, word_length);
+	}
 
 	correctly_spelled = enchant_result == 0;
 
@@ -418,6 +427,7 @@ gspell_checker_get_suggestions (GspellChecker *checker,
 				gssize         word_length)
 {
 	GspellCheckerPrivate *priv;
+	gchar *sanitized_word;
 	gchar **suggestions;
 	GSList *suggestions_list = NULL;
 	gint i;
@@ -433,7 +443,15 @@ gspell_checker_get_suggestions (GspellChecker *checker,
 		return NULL;
 	}
 
-	suggestions = enchant_dict_suggest (priv->dict, word, word_length, NULL);
+	if (_gspell_utils_str_to_ascii_apostrophe (word, word_length, &sanitized_word))
+	{
+		suggestions = enchant_dict_suggest (priv->dict, sanitized_word, -1, NULL);
+		g_free (sanitized_word);
+	}
+	else
+	{
+		suggestions = enchant_dict_suggest (priv->dict, word, word_length, NULL);
+	}
 
 	if (suggestions == NULL)
 	{
