@@ -27,8 +27,8 @@
 #include <string.h>
 #include <glib/gi18n-lib.h>
 #include "gspell-checker.h"
-#include "gspell-buffer-notifier.h"
 #include "gspell-text-buffer.h"
+#include "gspell-text-iter.h"
 #include "gspell-text-region.h"
 #include "gspell-utils.h"
 
@@ -70,6 +70,12 @@ enum
 	PROP_BUFFER,
 };
 
+typedef enum
+{
+	ADJUST_MODE_STRICTLY_INSIDE_WORD,
+	ADJUST_MODE_INCLUDE_NEIGHBORS,
+} AdjustMode;
+
 #define INLINE_CHECKER_TEXT_BUFFER_KEY	"GspellInlineCheckerTextBufferID"
 #define SUGGESTION_KEY			"GspellInlineSuggestionID"
 
@@ -96,8 +102,8 @@ check_word (GspellInlineCheckerTextBuffer *spell,
 		return;
 	}
 
-	if (!gtk_text_iter_starts_word (start) ||
-	    !gtk_text_iter_ends_word (end))
+	if (!_gspell_text_iter_starts_word (start) ||
+	    !_gspell_text_iter_ends_word (end))
 	{
 		g_warning ("Spell checking: @start and @end must delimit a word");
 		return;
@@ -126,6 +132,46 @@ check_word (GspellInlineCheckerTextBuffer *spell,
 	g_free (word);
 }
 
+static void
+adjust_iters (GtkTextIter *start,
+	      GtkTextIter *end,
+	      AdjustMode   mode)
+{
+	switch (mode)
+	{
+		case ADJUST_MODE_STRICTLY_INSIDE_WORD:
+			if (_gspell_text_iter_inside_word (start) &&
+			    !_gspell_text_iter_starts_word (start))
+			{
+				_gspell_text_iter_backward_word_start (start);
+			}
+
+			if (_gspell_text_iter_inside_word (end) &&
+			    !_gspell_text_iter_starts_word (end))
+			{
+				_gspell_text_iter_forward_word_end (end);
+			}
+			break;
+
+		case ADJUST_MODE_INCLUDE_NEIGHBORS:
+			if (_gspell_text_iter_ends_word (start) ||
+			    (_gspell_text_iter_inside_word (start) &&
+			     !_gspell_text_iter_starts_word (start)))
+			{
+				_gspell_text_iter_backward_word_start (start);
+			}
+
+			if (_gspell_text_iter_inside_word (end))
+			{
+				_gspell_text_iter_forward_word_end (end);
+			}
+			break;
+
+		default:
+			g_assert_not_reached ();
+	}
+}
+
 static void
 check_subregion (GspellInlineCheckerTextBuffer *spell,
 		 GtkTextIter                   *start,
@@ -135,17 +181,7 @@ check_subregion (GspellInlineCheckerTextBuffer *spell,
 
 	g_assert_cmpint (gtk_text_iter_compare (start, end), <=, 0);
 
-	if (gtk_text_iter_inside_word (start) &&
-	    !gtk_text_iter_starts_word (start))
-	{
-		gtk_text_iter_backward_word_start (start);
-	}
-
-	if (gtk_text_iter_inside_word (end) &&
-	    !gtk_text_iter_starts_word (end))
-	{
-		gtk_text_iter_forward_word_end (end);
-	}
+	adjust_iters (start, end, ADJUST_MODE_STRICTLY_INSIDE_WORD);
 
 	gtk_text_buffer_remove_tag (spell->buffer,
 				    spell->highlight_tag,
@@ -153,9 +189,9 @@ check_subregion (GspellInlineCheckerTextBuffer *spell,
 				    end);
 
 	word_start = *start;
-	if (!gtk_text_iter_starts_word (&word_start))
+	if (!_gspell_text_iter_starts_word (&word_start))
 	{
-		gtk_text_iter_forward_word_end (&word_start);
+		_gspell_text_iter_forward_word_end (&word_start);
 
 		/* Didn't move, there is no words after @start_adjusted. */
 		if (gtk_text_iter_equal (&word_start, start))
@@ -163,8 +199,8 @@ check_subregion (GspellInlineCheckerTextBuffer *spell,
 			return;
 		}
 
-		gtk_text_iter_backward_word_start (&word_start);
-		g_assert (gtk_text_iter_starts_word (&word_start));
+		_gspell_text_iter_backward_word_start (&word_start);
+		g_assert (_gspell_text_iter_starts_word (&word_start));
 		g_assert_cmpint (gtk_text_iter_compare (start, &word_start), <, 0);
 	}
 
@@ -174,18 +210,18 @@ check_subregion (GspellInlineCheckerTextBuffer *spell,
 		GtkTextIter word_end;
 		GtkTextIter next_word_start;
 
-		g_assert (gtk_text_iter_starts_word (&word_start));
+		g_assert (_gspell_text_iter_starts_word (&word_start));
 
 		word_end = word_start;
-		gtk_text_iter_forward_word_end (&word_end);
+		_gspell_text_iter_forward_word_end (&word_end);
 
 		g_assert_cmpint (gtk_text_iter_compare (&word_end, end), <=, 0);
 
 		check_word (spell, &word_start, &word_end);
 
 		next_word_start = word_end;
-		gtk_text_iter_forward_word_end (&next_word_start);
-		gtk_text_iter_backward_word_start (&next_word_start);
+		_gspell_text_iter_forward_word_end (&next_word_start);
+		_gspell_text_iter_backward_word_start (&next_word_start);
 
 		/* Make sure we've actually advanced (we don't advance if we
 		 * have just checked the last word of the buffer).
@@ -270,7 +306,6 @@ get_current_word_boundaries (GtkTextBuffer *buffer,
 			     GtkTextIter   *current_word_end)
 {
 	GtkTextIter insert_iter;
-	gboolean found = FALSE;
 
 	g_assert (current_word_start != NULL);
 	g_assert (current_word_end != NULL);
@@ -285,22 +320,12 @@ get_current_word_boundaries (GtkTextBuffer *buffer,
 					  gtk_text_buffer_get_insert (buffer));
 
 	*current_word_start = insert_iter;
-	if (gtk_text_iter_ends_word (current_word_start) ||
-	    (gtk_text_iter_inside_word (current_word_start) &&
-	     !gtk_text_iter_starts_word (current_word_start)))
-	{
-		gtk_text_iter_backward_word_start (current_word_start);
-		found = TRUE;
-	}
-
 	*current_word_end = insert_iter;
-	if (gtk_text_iter_inside_word (current_word_end))
-	{
-		gtk_text_iter_forward_word_end (current_word_end);
-		found = TRUE;
-	}
 
-	return found;
+	adjust_iters (current_word_start, current_word_end, ADJUST_MODE_INCLUDE_NEIGHBORS);
+
+	return (!gtk_text_iter_equal (current_word_start, &insert_iter) ||
+		!gtk_text_iter_equal (current_word_end, &insert_iter));
 }
 
 static void
@@ -353,13 +378,20 @@ check_visible_region_in_view (GspellInlineCheckerTextBuffer *spell,
 			_gspell_text_region_subtract (intersect,
 						      &current_word_start,
 						      &current_word_end);
+
+			/* Be sure that the current word will be re-checked
+			 * later when it will no longer be the current word.
+			 */
+			_gspell_text_region_add (spell->scan_region,
+						 &current_word_start,
+						 &current_word_end);
 		}
 	}
 
 	if (is_text_region_empty (intersect))
 	{
 		_gspell_text_region_destroy (intersect);
-		return;
+		goto out;
 	}
 
 	_gspell_text_region_get_iterator (intersect, &intersect_iter, 0);
@@ -368,6 +400,8 @@ check_visible_region_in_view (GspellInlineCheckerTextBuffer *spell,
 	{
 		GtkTextIter start;
 		GtkTextIter end;
+		GtkTextIter orig_start;
+		GtkTextIter orig_end;
 
 		if (!_gspell_text_region_iterator_get_subregion (&intersect_iter,
 								 &start,
@@ -376,8 +410,18 @@ check_visible_region_in_view (GspellInlineCheckerTextBuffer *spell,
 			break;
 		}
 
+		orig_start = start;
+		orig_end = end;
+
 		check_subregion (spell, &start, &end);
 
+		/* Ensure that we don't have an infinite loop. We must subtract
+		 * from scan_region at least [start, end], otherwise we will
+		 * re-check the same subregion again and again.
+		 */
+		g_assert (gtk_text_iter_compare (&start, &orig_start) <= 0);
+		g_assert (gtk_text_iter_compare (&orig_end, &end) <= 0);
+
 		_gspell_text_region_subtract (spell->scan_region, &start, &end);
 
 		_gspell_text_region_iterator_next (&intersect_iter);
@@ -391,6 +435,7 @@ check_visible_region_in_view (GspellInlineCheckerTextBuffer *spell,
 		spell->scan_region = NULL;
 	}
 
+out:
 	if (view != NULL)
 	{
 		/* FIXME properly. Workaround for bug in GtkTextView:
@@ -480,6 +525,34 @@ recheck_all (GspellInlineCheckerTextBuffer *spell)
 	check_visible_region (spell);
 }
 
+/* The word boundaries are not necessarily the same before and after a text
+ * insertion or deletion. We need the broader boundaries, so we connect to the
+ * signal without and with the AFTER flag.
+ */
+static void
+insert_text_before_cb (GtkTextBuffer                 *buffer,
+		       GtkTextIter                   *location,
+		       gchar                         *text,
+		       gint                           length,
+		       GspellInlineCheckerTextBuffer *spell)
+{
+	GtkTextIter start;
+	GtkTextIter end;
+
+	start = *location;
+	end = *location;
+	adjust_iters (&start, &end, ADJUST_MODE_INCLUDE_NEIGHBORS);
+	add_subregion_to_scan (spell, &start, &end);
+
+	/* Don't install_timeout(), it will anyway be called in
+	 * insert_text_after_cb(). If install_timeout() is called here, it is a
+	 * problem for the unit test mode because the subregion would be scanned
+	 * directly, but we need to wait that the text is inserted, otherwise
+	 * this can give different results (since the word boundaries are not
+	 * necessarily the same).
+	 */
+}
+
 static void
 insert_text_after_cb (GtkTextBuffer                 *buffer,
 		      GtkTextIter                   *location,
@@ -487,31 +560,121 @@ insert_text_after_cb (GtkTextBuffer                 *buffer,
 		      gint                           length,
 		      GspellInlineCheckerTextBuffer *spell)
 {
+	glong n_chars;
 	GtkTextIter start;
 	GtkTextIter end;
 
-	start = end = *location;
-	gtk_text_iter_backward_chars (&start, g_utf8_strlen (text, length));
+	n_chars = g_utf8_strlen (text, length);
 
-	/* Include neighbor words */
+	start = *location;
+	end = *location;
+	gtk_text_iter_backward_chars (&start, n_chars);
 
-	if (gtk_text_iter_ends_word (&start) ||
-	    (gtk_text_iter_inside_word (&start) &&
-	     !gtk_text_iter_starts_word (&start)))
-	{
-		gtk_text_iter_backward_word_start (&start);
-	}
-
-	if (gtk_text_iter_inside_word (&end))
-	{
-		gtk_text_iter_forward_word_end (&end);
-	}
-
-	spell->check_current_word = FALSE;
+	adjust_iters (&start, &end, ADJUST_MODE_INCLUDE_NEIGHBORS);
 	add_subregion_to_scan (spell, &start, &end);
+
+	/* Check current word? */
+
+	/* If more than one character is inserted, it's probably not a normal
+	 * keypress, e.g. a clipboard paste or DND. So it's better to check the
+	 * current word in that case, to know ASAP if the word is correctly
+	 * spelled.
+	 * The same if e.g. a space or punctuation is inserted: in that case we
+	 * are not editing the current word. Maybe a word has been split in two,
+	 * in which case the word on the left will anyway be checked, so it's
+	 * better to know directly whether the word on the right is correctly
+	 * spelled as well, so we know if we need to edit it or not.
+	 * And if there is a selection, it means that the text was inserted
+	 * programmatically, so the user is not editing the current word
+	 * manually.
+	 */
+	if (n_chars > 1 ||
+	    !g_unichar_isalnum (g_utf8_get_char (text)) ||
+	    gtk_text_buffer_get_has_selection (buffer))
+	{
+		spell->check_current_word = TRUE;
+	}
+	else
+	{
+		GtkTextIter cursor_pos;
+
+		gtk_text_buffer_get_iter_at_mark (buffer,
+						  &cursor_pos,
+						  gtk_text_buffer_get_insert (buffer));
+
+		spell->check_current_word = !gtk_text_iter_equal (location, &cursor_pos);
+	}
+
 	install_timeout (spell, TIMEOUT_DURATION_BUFFER_MODIFIED);
 }
 
+/* Same reasoning as for the ::insert-text signal. */
+static void
+delete_range_before_cb (GtkTextBuffer                 *buffer,
+			GtkTextIter                   *start,
+			GtkTextIter                   *end,
+			GspellInlineCheckerTextBuffer *spell)
+{
+	GtkTextIter start_adjusted;
+	GtkTextIter end_adjusted;
+
+	start_adjusted = *start;
+	end_adjusted = *end;
+	adjust_iters (&start_adjusted, &end_adjusted, ADJUST_MODE_INCLUDE_NEIGHBORS);
+	add_subregion_to_scan (spell, &start_adjusted, &end_adjusted);
+
+	/* Check current word? */
+	if (gtk_text_buffer_get_has_selection (buffer) ||
+	    gtk_text_iter_get_line (start) != gtk_text_iter_get_line (end) ||
+	    (gtk_text_iter_get_line_offset (end) - gtk_text_iter_get_line_offset (start)) > 1)
+	{
+		spell->check_current_word = TRUE;
+	}
+	else
+	{
+		GtkTextIter cursor_pos;
+		gboolean is_backspace;
+		gboolean is_delete;
+
+		gtk_text_buffer_get_iter_at_mark (buffer,
+						  &cursor_pos,
+						  gtk_text_buffer_get_insert (buffer));
+
+		is_backspace = gtk_text_iter_equal (&cursor_pos, end);
+		is_delete = gtk_text_iter_equal (&cursor_pos, start);
+
+		if (is_backspace)
+		{
+			if (_gspell_text_iter_inside_word (start) ||
+			    _gspell_text_iter_ends_word (start))
+			{
+				spell->check_current_word = FALSE;
+			}
+			else
+			{
+				spell->check_current_word = TRUE;
+			}
+		}
+		else if (is_delete)
+		{
+			if (_gspell_text_iter_inside_word (end) ||
+			    _gspell_text_iter_ends_word (end))
+			{
+				spell->check_current_word = FALSE;
+			}
+			else
+			{
+				spell->check_current_word = TRUE;
+			}
+		}
+		/* Text deleted programmatically. */
+		else
+		{
+			spell->check_current_word = TRUE;
+		}
+	}
+}
+
 static void
 delete_range_after_cb (GtkTextBuffer                 *buffer,
 		       GtkTextIter                   *start,
@@ -521,26 +684,13 @@ delete_range_after_cb (GtkTextBuffer                 *buffer,
 	GtkTextIter start_adjusted;
 	GtkTextIter end_adjusted;
 
+	g_return_if_fail (gtk_text_iter_equal (start, end));
+
 	start_adjusted = *start;
 	end_adjusted = *end;
-
-	/* Just to be sure. Normally start == end. */
-	gtk_text_iter_order (&start_adjusted, &end_adjusted);
-
-	if (gtk_text_iter_ends_word (&start_adjusted) ||
-	    (gtk_text_iter_inside_word (&start_adjusted) &&
-	     !gtk_text_iter_starts_word (&start_adjusted)))
-	{
-		gtk_text_iter_backward_word_start (&start_adjusted);
-	}
-
-	if (gtk_text_iter_inside_word (&end_adjusted))
-	{
-		gtk_text_iter_forward_word_end (&end_adjusted);
-	}
-
-	spell->check_current_word = FALSE;
+	adjust_iters (&start_adjusted, &end_adjusted, ADJUST_MODE_INCLUDE_NEIGHBORS);
 	add_subregion_to_scan (spell, &start_adjusted, &end_adjusted);
+
 	install_timeout (spell, TIMEOUT_DURATION_BUFFER_MODIFIED);
 }
 
@@ -566,22 +716,22 @@ get_word_extents_at_click_position (GspellInlineCheckerTextBuffer *spell,
 
 	gtk_text_buffer_get_iter_at_mark (spell->buffer, &iter, spell->mark_click);
 
-	if (!gtk_text_iter_inside_word (&iter) &&
-	    !gtk_text_iter_ends_word (&iter))
+	if (!_gspell_text_iter_inside_word (&iter) &&
+	    !_gspell_text_iter_ends_word (&iter))
 	{
 		return FALSE;
 	}
 
 	*start = iter;
-	if (!gtk_text_iter_starts_word (start))
+	if (!_gspell_text_iter_starts_word (start))
 	{
-		gtk_text_iter_backward_word_start (start);
+		_gspell_text_iter_backward_word_start (start);
 	}
 
 	*end = iter;
-	if (!gtk_text_iter_ends_word (end))
+	if (!_gspell_text_iter_ends_word (end))
 	{
-		gtk_text_iter_forward_word_end (end);
+		_gspell_text_iter_forward_word_end (end);
 	}
 
 	return TRUE;
@@ -844,8 +994,8 @@ remove_tag_to_word (GspellInlineCheckerTextBuffer *spell,
 			break;
 		}
 
-		if (gtk_text_iter_starts_word (&match_start) &&
-		    gtk_text_iter_ends_word (&match_end))
+		if (_gspell_text_iter_starts_word (&match_start) &&
+		    _gspell_text_iter_ends_word (&match_end))
 		{
 			gtk_text_buffer_remove_tag (spell->buffer,
 						    spell->highlight_tag,
@@ -1021,16 +1171,16 @@ set_spell_checker (GspellInlineCheckerTextBuffer *spell,
 }
 
 static void
-text_buffer_checker_changed_cb (GspellBufferNotifier          *notifier,
-				GtkTextBuffer                 *buffer,
-				GspellChecker                 *new_checker,
-				GspellInlineCheckerTextBuffer *spell)
+spell_checker_notify_cb (GspellTextBuffer              *gspell_buffer,
+			 GParamSpec                    *pspec,
+			 GspellInlineCheckerTextBuffer *spell)
 {
-	if (spell->buffer == buffer)
-	{
-		set_spell_checker (spell, new_checker);
-		recheck_all (spell);
-	}
+	GspellChecker *new_checker;
+
+	new_checker = gspell_text_buffer_get_spell_checker (gspell_buffer);
+	set_spell_checker (spell, new_checker);
+
+	recheck_all (spell);
 }
 
 static void
@@ -1039,8 +1189,8 @@ set_buffer (GspellInlineCheckerTextBuffer *spell,
 {
 	GtkTextTagTable *tag_table;
 	GtkTextIter start;
+	GspellTextBuffer *gspell_buffer;
 	GspellChecker *checker;
-	GspellBufferNotifier *notifier;
 
 	g_return_if_fail (GTK_IS_TEXT_BUFFER (buffer));
 	g_return_if_fail (spell->buffer == NULL);
@@ -1054,12 +1204,24 @@ set_buffer (GspellInlineCheckerTextBuffer *spell,
 			   INLINE_CHECKER_TEXT_BUFFER_KEY,
 			   spell);
 
+	g_signal_connect_object (buffer,
+				 "insert-text",
+				 G_CALLBACK (insert_text_before_cb),
+				 spell,
+				 0);
+
 	g_signal_connect_object (buffer,
 				 "insert-text",
 				 G_CALLBACK (insert_text_after_cb),
 				 spell,
 				 G_CONNECT_AFTER);
 
+	g_signal_connect_object (buffer,
+				 "delete-range",
+				 G_CALLBACK (delete_range_before_cb),
+				 spell,
+				 0);
+
 	g_signal_connect_object (buffer,
 				 "delete-range",
 				 G_CALLBACK (delete_range_after_cb),
@@ -1115,13 +1277,13 @@ set_buffer (GspellInlineCheckerTextBuffer *spell,
 	gtk_text_buffer_get_start_iter (spell->buffer, &start);
 	spell->mark_click = gtk_text_buffer_create_mark (spell->buffer, NULL, &start, TRUE);
 
-	checker = gspell_text_buffer_get_spell_checker (spell->buffer);
+	gspell_buffer = gspell_text_buffer_get_from_gtk_text_buffer (spell->buffer);
+	checker = gspell_text_buffer_get_spell_checker (gspell_buffer);
 	set_spell_checker (spell, checker);
 
-	notifier = _gspell_buffer_notifier_get_instance ();
-	g_signal_connect_object (notifier,
-				 "text-buffer-checker-changed",
-				 G_CALLBACK (text_buffer_checker_changed_cb),
+	g_signal_connect_object (gspell_buffer,
+				 "notify::spell-checker",
+				 G_CALLBACK (spell_checker_notify_cb),
 				 spell,
 				 0);
 
